'''
编译命令
python -m grpc_tools.protoc -I. --python_out=. --grpc_python_out=. ./rpc_data.proto
'''
import json
import os
import threading
import time
import requests

from hs_rpc.exceptions import RpcFuncRegisterError
from hs_rpc.request import rpc_request_invoke


os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"

def rpc_server_invoke(request):
    """
    服务端调用入口
    :param request: 客户端传入的请求对象
    :return: 响应
    """
    request_data = json.loads(request)
    rule = request_data.pop('func')
    response = rpc.dispatch(rule, **request_data)

    return json.dumps(response.__dict__)


class Response:
    def __init__(self, message, code):
        self.message = message
        self.code = code


class RpcServerHelper:

    def __init__(self):
        self.app = None
        self.rpc_map = {}
        self.response_class = Response

    def rpc(self, rule=None, token_verify=True):
        """
        rpc函数注册器
        :param rule: rpc函数注册名
        :param token_verify: 是否需要验证token才能访问
        :return:
        """
        def decorator(rpc_func):
            self.add_rpc_rule(rule, rpc_func, token_verify)
            return rpc_func

        return decorator

    def add_rpc_rule(self, rule, rpc_func, token_verify):
        """
        注册rpc函数到映射器
        :param rule: rpc函数注册名
        :param rpc_func: 装饰的函数
        :param token_verify: 是否需要验证token才能访问
        """
        if rule is None:
            rule = self._endpoint_from_rpc_func(rpc_func)

        if rpc_func is not None:
            old_func_obj = self.rpc_map.get(rule)
            if old_func_obj is not None:
                old_func = old_func_obj['func']
                if old_func != rpc_func:
                    raise RpcFuncRegisterError('rpc函数注册规则重复【%s】' % rule)
            self.rpc_map[rule] = {'func':rpc_func, 'token_verify':token_verify}

    def _endpoint_from_rpc_func(self, rpc_func):
        """
        如果不提供端点，则使用函数名作为key注册
        :param rpc_func: 提供注册为rpc服务函数的注册名
        :return:
        """
        assert rpc_func is not None, 'expected rpc_func if endpoint ' \
                                     'is not provided.'
        return rpc_func.__name__

    def _verify_token(self, token):
        """
        验证token，此时又可以作为客户端访问。。。。。
        :param token:
        :return:
        """
        response = rpc_request_invoke(func='token_verify', message=token, app='HSRIGHT')
        if response['code'] != 200:
            return {'message':response['message'],
                    'code':response['code']}

    def dispatch(self, rule, **options):
        func =  self.rpc_map[rule]
        if func['token_verify']:
            token = options.pop('token')
            if not token:
                return {
                    'message':'未传入token',
                    'code':401
                }
            response = self._verify_token(token)
            if response is not None:
                return response
        return self.make_response(func['func'](**options))

    def make_response(self, rv):
        if isinstance(rv, tuple):
            if len(rv) == 2:
                message, code = rv
                return self.response_class(message=message, code=code)
            else:
                raise TypeError('请返回一个有效元组，(message, code)')

        if rv is None:
            raise TypeError('rpc函数必须返回一个有效的响应')

        if not isinstance(rv, self.response_class):
            raise TypeError('rpc函数必须返回一个Response实例，或者元组(message, code)')

        return rv

    def run(self, app):
        self.app = app

        # 在flask启动服务时来激活rpc服务
        @app.before_first_request
        def activate_rpc():
            from hs_rpc.server import serve
            t = threading.Thread(target=serve, args=(app,))
            t.start()

        # 循环去激活activate_rpc执行，这里会稍微有些绕
        def start_loop():
            not_started = True
            while not_started:
                try:
                    r = requests.get('http://localhost:{}/'.format(int(app.config.get('IP_PORT')) or 5000))
                    if r.status_code == 200:
                        self.app.logger.info('rpc服务已启动成功！！！')
                        not_started = False
                except:
                    self.app.logger.info('rpc服务启动中...')
                time.sleep(2)

        app.logger.info('正在激活rpc服务...')
        t = threading.Thread(target=start_loop)
        t.start()



rpc = RpcServerHelper()


'''
编译命令
python -m grpc_tools.protoc -I. --python_out=. --grpc_python_out=. ./rpc_data.proto
'''
import json
import os

import grpc
from flask import current_app

from hs_rpc.exceptions import MissingConfigError
from hs_rpc.proto.rpc_data_pb2 import Request
from hs_rpc.proto.rpc_data_pb2_grpc import GreeterStub

os.environ["PROTOCOL_BUFFERS_PYTHON_IMPLEMENTATION"] = "python"


def rpc_request_invoke(func, message, app):
    '''
    调用指定app的rpc服务方的一个函数
    :param func: 服务方的函数名
    :param message: 服务方函数需要的参数，字典类型，如果服务方需要token验证message必须携带token
    :param app: 将要调用哪个app的rpc服务
    :return:
    '''
    app_rpc = current_app.config.get((app + '_' + 'rpc').upper())
    if app_rpc is None:
        raise MissingConfigError('未配置 %s 的服务地址' % app)
    if not isinstance(message, dict):
        raise MissingConfigError('message类型必须为字典类型')

    message.update({'func': func})
    channel = grpc.insecure_channel(app_rpc)
    stub = GreeterStub(channel)
    response = stub.handle_request(Request(message=json.dumps(message, ensure_ascii=True)))
    return json.loads(response.message)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc
from . import rpc_data_pb2 as rpc__data__pb2


class GreeterStub(object):
  """The greeting service definition.
  """

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.handle_request = channel.unary_unary(
        '/rpc_data.Greeter/handle_request',
        request_serializer=rpc__data__pb2.Request.SerializeToString,
        response_deserializer=rpc__data__pb2.Reply.FromString,
        )


class GreeterServicer(object):
  """The greeting service definition.
  """

  def handle_request(self, request, context):
    """Sends a greeting
    """
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_GreeterServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'handle_request': grpc.unary_unary_rpc_method_handler(
          servicer.handle_request,
          request_deserializer=rpc__data__pb2.Request.FromString,
          response_serializer=rpc__data__pb2.Reply.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'rpc_data.Greeter', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



import grpc
import time
from concurrent import futures

from hs_rpc.proto import rpc_data_pb2_grpc
from hs_rpc.proto.rpc_data_pb2 import Reply
from hs_rpc.proto.rpc_data_pb2_grpc import add_GreeterServicer_to_server
from hs_rpc.rpc import rpc_server_invoke

_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class Greeter(rpc_data_pb2_grpc.GreeterServicer):

    def handle_request(self, request, context):
        response = rpc_server_invoke(request.message)
        return Reply(message=response)


def serve(app):
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    add_GreeterServicer_to_server(Greeter(), server)

    server.add_insecure_port('[::]:{}'.format(app.config['RPC_PORT'] or 8120))
    server.start()
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hs_rpc",
    version="0.0.8",
    author="huansi_mes_team",
    author_email="751995207@qq.com",
    description="rpc服务打包",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'Click',
        'Flask',
        'google',
        'grpcio',
        'protobuf',
        'soupsieve',
        'six',
    ],
)